(* -*- coq-prog-args: ("-compat" "8.15") -*- *)
(** Check that the current-minus-two compatibility flag actually requires the relevant modules. *)
Import Coq.Compat.Coq817.
Import Coq.Compat.Coq816.
Import Coq.Compat.Coq815.
