overlay elpi https://github.com/gares/coq-elpi dynlink-w-findlib 15220

overlay equations https://github.com/gares/Coq-Equations dynlink-w-findlib 15220

overlay rewriter https://github.com/gares/rewriter dynlink-w-findlib 15220

overlay bignums https://github.com/gares/bignums dynlink-w-findlib 15220

overlay stdlib2 https://github.com/gares/stdlib2 dynlink-w-findlib 15220

overlay metacoq https://github.com/gares/metacoq dynlink-w-findlib 15220

# overlay fiat_crypto https://github.com/gares/fiat-crypto dynlink-w-findlib 15220

overlay simple_io https://github.com/gares/coq-simple-io dynlink-w-findlib 15220

overlay quickchick https://github.com/gares/QuickChick dynlink-w-findlib 15220

overlay paramcoq https://github.com/gares/paramcoq dynlink-w-findlib 15220

overlay aac_tactics https://github.com/gares/aac-tactics dynlink-w-findlib 15220

overlay mtac2 https://github.com/gares/Mtac2 dynlink-w-findlib 15220

overlay unicoq https://github.com/gares/unicoq dynlink-w-findlib 15220

overlay relation_algebra https://github.com/gares/relation-algebra dynlink-w-findlib 15220

overlay coqhammer https://github.com/gares/coqhammer dynlink-w-findlib 15220

overlay coq_dpdgraph https://github.com/gares/coq-dpdgraph dynlink-w-findlib 15220

overlay itauto https://gitlab.inria.fr/tassi/itauto dynlink-w-findlib 15220

overlay fiat_parsers https://github.com/gares/fiat dynlink-w-findlib 15220

overlay reduction_effects https://github.com/gares/reduction-effects dynlink-w-findlib 15220

# overlay hott https://github.com/gares/HoTT dynlink-w-findlib 15220

overlay coq_performance_tests https://github.com/gares/coq-performance-tests dynlink-w-findlib 15220

overlay coq_tools https://github.com/gares/coq-tools dynlink-w-findlib 15220
