overlay elpi https://github.com/ppedrot/coq-elpi universes-rm-univ-of-sort 15856

overlay equations https://github.com/ppedrot/Coq-Equations universes-rm-univ-of-sort 15856

overlay itauto https://gitlab.inria.fr/pedrot/itauto universes-rm-univ-of-sort 15856

overlay metacoq https://github.com/ppedrot/metacoq universes-rm-univ-of-sort 15856

overlay unicoq https://github.com/ppedrot/unicoq universes-rm-univ-of-sort 15856
