overlay equations https://github.com/ppedrot/Coq-Equations deprecate-sigma-wrapper 15393

overlay aac_tactics https://github.com/ppedrot/aac-tactics deprecate-sigma-wrapper 15393

overlay itauto https://gitlab.inria.fr/pedrot/itauto deprecate-sigma-wrapper 15393
