overlay equations https://github.com/ppedrot/Coq-Equations evar-compact-instance 16442

overlay metacoq https://github.com/ppedrot/metacoq evar-compact-instance 16442

overlay quickchick https://github.com/ppedrot/QuickChick evar-compact-instance 16442

overlay elpi https://github.com/ppedrot/coq-elpi evar-compact-instance 16442

overlay coqhammer https://github.com/ppedrot/coqhammer evar-compact-instance 16442

overlay unicoq https://github.com/ppedrot/unicoq evar-compact-instance 16442

overlay mtac2 https://github.com/ppedrot/Mtac2 evar-compact-instance 16442
