overlay equations https://github.com/herbelin/Coq-Equations coq-master+adapt-coq-pr15344-renaming-syntax_def.ml-abbreviation.ml 15344 master+renaming-syntax_def.ml-abbreviation.ml
overlay elpi https://github.com/herbelin/coq-elpi coq-master+adapt-coq-pr15344-renaming-syntax_def.ml-abbreviation.ml 15344 master+renaming-syntax_def.ml-abbreviation.ml
