#overlay <project> <giturl> <ref> <prnumber> [<prbranch>]

#overlay JasonGross/coq-tools

#JasonGross/coq-tools#117

overlay coq_tools https://github.com/JasonGross/coq-tools fix-keep-error 15880 coq_mk-delete-on-err
