Declare Custom Entry foo.

Print Custom Grammar foo.

Notation "[ e ]" := e (e custom foo at level 0).

Print Custom Grammar foo.

Notation "1" := O (in custom foo at level 0).

Print Custom Grammar foo.

Fail Declare Custom Entry foo.
