overlay hott https://github.com/herbelin/HoTT master+adapt-coq-pr15420-fix-naming-arguments-ind-schemes 15537 master+fix15420-hardwire-names-pred-mutual-induction-schemes
overlay stdlib2 https://github.com/herbelin/stdlib2 master+adapt-coq-pr15420-fix-naming-arguments-ind-schemes 15537 master+fix15420-hardwire-names-pred-mutual-induction-schemes
overlay mtac2 https://github.com/herbelin/Mtac2 master+adapt-coq-pr15420-fix-naming-arguments-ind-schemes 15537 master+fix15420-hardwire-names-pred-mutual-induction-schemes
