overlay elpi https://github.com/ppedrot/coq-elpi reduce-universe-footprint 15837

overlay metacoq https://github.com/ppedrot/metacoq reduce-universe-footprint 15837

overlay mtac2 https://github.com/ppedrot/Mtac2 reduce-universe-footprint 15837

overlay unicoq https://github.com/ppedrot/unicoq reduce-universe-footprint 15837
