overlay elpi https://github.com/SkySkimmer/coq-elpi libobject-names 15317

overlay equations https://github.com/SkySkimmer/Coq-Equations libobject-names 15317

overlay paramcoq https://github.com/SkySkimmer/paramcoq libobject-names 15317

overlay quickchick https://github.com/SkySkimmer/QuickChick libobject-names 15317
