overlay unicoq https://github.com/ppedrot/unicoq stricter-evar-access 16488

overlay rewriter https://github.com/ppedrot/rewriter stricter-evar-access 16488

overlay equations https://github.com/ppedrot/Coq-Equations stricter-evar-access 16488

overlay mtac2 https://github.com/ppedrot/Mtac2 stricter-evar-access 16488

overlay relation_algebra https://github.com/ppedrot/relation-algebra stricter-evar-access 16488
