Fail Test Search Blacklists.
Fail Test Search Blacklist for foo.
Set Search Blacklists.
Fail Remove Search Blacklists "bar" foo.
Fail Remove Search Blacklist "bar" foo.
Fail Add Search Blacklist "bar" foo.
