
.node-unpublished {
  background-color: #fff4f4;
}
.preview .node {
  background-color: #ffffea;
}
#node-admin-filter ul {
  list-style-type: none;
  padding: 0;
  margin: 0;
  width: 100%;
}
#node-admin-buttons {
  float: left; /* LTR */
  margin-left: 0.5em; /* LTR */
  clear: right; /* LTR */
}
td.revision-current {
  background: #ffc;
}
.node-form .form-text {
  display: block;
  width: 95%;
}
.node-form .container-inline .form-text {
  display: inline;
  width: auto;
}
.node-form .standard {
  clear: both;
}
.node-form textarea {
  display: block;
  width: 95%;
}
.node-form .attachments fieldset {
  float: none;
  display: block;
}
.terms-inline {
  display: inline;
}
