overlay equations https://github.com/gares/Coq-Equations syndef-principal-scope 13965
overlay elpi https://github.com/gares/coq-elpi syndef-principal-scope 13965
overlay paramcoq https://github.com/gares/paramcoq syndef-principal-scope 13965
