overlay aac_tactics https://github.com/ppedrot/aac-tactics split-tclfail-level 15327

overlay coqhammer https://github.com/ppedrot/coqhammer split-tclfail-level 15327

overlay elpi https://github.com/ppedrot/coq-elpi split-tclfail-level 15327

overlay equations https://github.com/ppedrot/Coq-Equations split-tclfail-level 15327

overlay fiat_parsers https://github.com/ppedrot/fiat split-tclfail-level 15327

overlay itauto https://gitlab.inria.fr/pedrot/itauto split-tclfail-level 15327

overlay unicoq https://github.com/ppedrot/unicoq split-tclfail-level 15327

overlay relation_algebra https://github.com/ppedrot/relation-algebra split-tclfail-level 15327
